<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8">
    <title>{{ name }}</title>
    <style>
      h1 {
        margin-bottom: 1em;
      }
    </style>
  </head>
  <body>
    {% if path %}
    <h1>{% trans %}Taguette highlights: {{ path }}{% endtrans %}</h1>
    {% else %}
    <h1>{% trans %}Taguette highlights{% endtrans %}</h1>
    {% endif %}
{% for hl in highlights %}
    {{ hl.snippet |safe }}
    <p>
      {% trans doc=hl.document.name %}<strong>Document:</strong> {{ doc }}{% endtrans %}
      <strong>{% trans %}Tags:{% endtrans %}</strong>
      {% for tag in hl.tags -%}
        {{ tag.path }}
        {%- if not loop.last %}, {% endif %}
      {% endfor %}
    </p>
    {% if not loop.last %}<hr>{% endif %}
{% endfor %}
  </body>
</html>
